-- global.lua
-- Author: Vicent Gong
-- Date: 2012-09-30
-- Last modification : 2013-05-30
-- Description: provide some global functions

FwLog = function(logStr)
	print_string(logStr);
end

Copy = function(t)
	if type(t) ~= "table" then
		return t;
	end

	local ret = {};
	for k,v in pairs(t) do 
		ret[Copy(k)] =Copy(v);
	end

    local mt = getmetatable(t);
    setmetatable(ret,Copy(mt));
	return ret;
end

MegerTables = function(...)
    local ret = {};
    local count = select("#",...);
    for i=1,count do
        local t = select(i,...);
        if type(t) == "table" then
            for _,v in pairs(t) do
                ret[#ret+1] = v;
            end
        end
    end
    return ret;
end

CreateTable = function(weakOption)
    if not weakOption or (weakOption ~= "k" and weakOption ~= "v" and weakOption ~= "kv") then
        return {};
    end

    local ret = {};
    setmetatable(ret,{__mode = weakOption});
    return ret;
end

GetNumFromJsonTable = function(tb, key, default)
    local ret = default or 0;
    if tb[key] ~= nil then
        if tb[key]:get_value() ~= nil then
            ret = tonumber(tb[key]:get_value());
            if ret == nil then
                ret = default or -1;
            end
        end
    end
    return ret;
end

GetStrFromJsonTable = function(tb, key, default)
    local ret = default or "";
    if tb[key] ~= nil then
        if tb[key]:get_value() ~= nil then
            ret = tb[key]:get_value() or "";
            if string.len(ret)  == 0 then
                ret = default;
            end
        end
    end
    return ret;
end

GetBlooeanFromJsonTable = function(tb, key, default)
    local ret = default;
    if tb and tb[key] ~= nil then
        if tb[key]:get_value() ~= nil then
            ret = tb[key]:get_value();
        end
    end
    return ret;
end

GetTableFromJsonTable = function(tb, key, default)
    local ret = default;
    if tb and tb[key] ~= nil then
        if tb[key]:get_value() ~= nil then
            ret = tb[key]:get_value();
            if type(ret) ~= "table" then
                ret = default;
            end
        end
    end
    return ret;
end

orderedPairs = function(t)
	local cmpMultitype = function(op1, op2)
		local type1, type2 = type(op1), type(op2)
		if type1 ~= type2 then --cmp by type
			return type1 < type2
		elseif type1 == "number" and type2 == "number"
			or type1 == "string" and type2 == "string" then
			return op1 < op2 --comp by default
		elseif type1 == "boolean" and type2 == "boolean" then
			return op1 == true
		else
			return tostring(op1) < tostring(op2) --cmp by address
		end
	end

	local genOrderedIndex = function(t)
		local orderedIndex = {}
		for key in pairs(t) do
			table.insert( orderedIndex, key )
		end
		table.sort( orderedIndex, cmpMultitype ) --### CANGE ###
		return orderedIndex
	end

    local orderedIndex = genOrderedIndex( t );
    local i = 0;
    return function(t)
        i = i + 1;
        if orderedIndex[i] then
            return orderedIndex[i],t[orderedIndex[i]];
        end
    end,t, nil;
end


Joins = function(t, mtkey)
    local str = "K";
    if t == nil or type(t) == "boolean"  or type(t) == "byte" then
        return str;
    elseif type(t) == "number" or type(t) == "string" then
        str = string.format("%sT%s%s", str.."", mtkey, string.gsub(t, "[^a-zA-Z0-9]",""));
    elseif type(t) == "table" then
        for k,v in orderedPairs(t) do
            str = string.format("%s%s=%s", str, tostring(k), Joins(v, mtkey));
        end
    end
    return str;
end

--[[--

输出值的内容

### 用法示例

~~~ lua

local t = {comp = "chukong", engine = "quick"}

dump(t)

~~~

@param mixed value 要输出的值

@param [string desciption] 输出内容前的文字描述

@parma [integer nesting] 输出时的嵌套层级，默认为 3

]]
--[===[
dump = function(value, desciption, nesting)
    if type(nesting) ~= "number" then nesting = 3 end

    local lookupTable = {}
    local result = {}

    local function _v(v)
        if type(v) == "string" then
            v = "\"" .. v .. "\""
        end
        return tostring(v)
    end

    local traceback = string.split(debug.traceback("", 2), "\n")
    print("dump from: " .. string.trim(traceback[3]))

    local function _dump(value, desciption, indent, nest, keylen)
        desciption = desciption or "<var>"
        spc = ""
        if type(keylen) == "number" then
            spc = string.rep(" ", keylen - string.len(_v(desciption)))
        end
        if type(value) ~= "table" then
            result[#result +1 ] = string.format("%s%s%s = %s", indent, _v(desciption), spc, _v(value))
        elseif lookupTable[value] then
            result[#result +1 ] = string.format("%s%s%s = *REF*", indent, desciption, spc)
        else
            lookupTable[value] = true
            if nest > nesting then
                result[#result +1 ] = string.format("%s%s = *MAX NESTING*", indent, desciption)
            else
                result[#result +1 ] = string.format("%s%s = {", indent, _v(desciption))
                local indent2 = indent.."    "
                local keys = {}
                local keylen = 0
                local values = {}
                for k, v in pairs(value) do
                    keys[#keys + 1] = k
                    local vk = _v(k)
                    local vkl = string.len(vk)
                    if vkl > keylen then keylen = vkl end
                    values[k] = v
                end
                table.sort(keys, function(a, b)
                    if type(a) == "number" and type(b) == "number" then
                        return a < b
                    else
                        return tostring(a) < tostring(b)
                    end
                end)
                for i, k in ipairs(keys) do
                    _dump(values[k], k, indent2, nest + 1, keylen)
                end
                result[#result +1] = string.format("%s}", indent)
            end
        end
    end
    _dump(value, desciption, "- ", 1)

    for i, line in ipairs(result) do
        print(line)
    end
end
--]===]

